/* YOODA : Yet another Object Oriented Database
   Copyright (C) 1994  Eric Abecassis

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */


/*****************************************************************************
******************************************************************************
**                                                                            
**        Y     Y   OOOO    OOOO   DDDDD     AA
**         Y   Y   O    O  O    O  D    D   A  A
**          Y Y    O    O  O    O  D    D  A    A 
**           Y     O    O  O    O  D    D  AAAAAA
**           Y     O    O  O    O  D    D  A    A
**           Y      OOOO    OOOO   DDDDD   A    A
**
******************************************************************************
******************************************************************************/
/*****************************************************************************
*
*  MODULE :             Cluster management
*  SCOPE :              Cluster management
*  FILE :               cluster.cc
*  AUTHORS :            E. Abecassis
*  CREATION DATE :      August 1, 1993
*  COPYRIGHTS :        (C) E. Abecassis, 1993
*  HISTORY :
*
******************************************************************************/

#ifdef __GNUG__
#pragma implementation
#endif

/* IMPORTS *******************************************************************/

#ifndef __cluster_HXX
#include "cluster.hxx"
#endif

#ifndef __pagehdlmem_HXX
#include "pagehdlmem.hxx"
#endif

#ifndef __define_H
#include "define.h"
#endif

#ifndef __page_HXX
#include "page.hxx"
#endif

#ifndef __virtualdbserver_HXX
#include "virtualdbserver.hxx"
#endif

#ifndef __yoofunc_HXX
#include "yoofunc.hxx"
#endif

#ifndef __yooexcept_HXX
#include "yooexcept.hxx"
#endif

#ifndef __smartp_HXX
#include "smartp.hxx"
#endif

#ifndef __var_H
#include "var.h"
#endif


/*****************************************************************************/


ClusterSet Cluster::AllocSet ;


/*= EXTERNAL FUNCTION =======================================================*/
void BlocAllocator::insert(PageHdlMem* pd)  
/*=
[DESCRIPTION :]
  Insert a page in the allocator. The free space is computed from usedSpace 
  member. A test is done to insert at the head of the list, the page which 
  already have an exclusive lock and at the tail, the page with share lock.
  So when a space is allocated, page with exclusive lock are choose first.
  Note that the allocator structure contains only short page.

  (IN)
    <PageHdlMem* pd>
      => Page descriptor of the corresponding page

[DIAGNOSTIC :]
  <YOO_ERR_DBG>
    => error if the page seems to be inconsistent

[CALL :]
  Cluster 
    => Call when a page is inserted in the cluster
=*/
{
#ifdef DEBUG
  if (pd->consistency() == FALSE) {
    yoo_error(YOO_ERR_DBG, "BlocAlloc::insert => Strange page") ;
    yoo_exit(YOO_ERR_DBG) ;
  }
#endif
  
  if (pd->freeSpace() == 0) return ;
  
  if (pd->getLock() <= YOO_SHARE) {
    // insert at the tail
    this->_freeList[pd->freeSpace()-1].insertTail(pd) ;
  } else {
    // insert at the head
    this->_freeList[pd->freeSpace()-1].insertHead(pd) ;
  }
}


/*= EXTERNAL FUNCTION =======================================================*/
ObjectX* BlocAllocator::alloc(int sizeSlot)  
/*=
[DESCRIPTION :]
  Find a page to allocate a block of size 'sizeslot'. Remember that page are 
  inserted by using 'usedSpace' value of the page. So it is possible that 
  free space is not contiguous and then allocation could fail. In such case,
  the page is reinserted in the previous entry of the allocation table.
  
  (IN)
    <int [1..SLOT_IN_PAGE] sizeSlot>
      => Size in slot of the desired block.

  (OUT)
    <ObjectX*>
      => Virtual memory pointer on allocated object. The object is not 
      => initialised.

[DIAGNOSTIC :]
  (RETURN)
    <NULL>
      => Allocation failed. The cluster will allocate more free pages
    <ObjectX*>
      => Allocated object uninitialised

[CALL :]
  Cluster
    => Call when allocation is asked on the cluster
=*/
{
#ifdef DEBUG 
  if ((sizeSlot > SLOT_IN_PG) || (sizeSlot <= 0)) {
    yoo_error(YOO_ERR_DBG, "=> Strange size for an object to alloc") ;
    yoo_exit(YOO_ERR_DBG) ;
  }
#endif

  ObjectX* res = NULL ;
  PageHdlMem* pd = NULL ;

  // Scan list of pages which can allocate space.
  // Search is done as a best fit search.
  for (int i = sizeSlot-1; i < SLOT_IN_PG; i++) {
    while ((pd = (PageHdlMem*)this->_freeList[i].head()) != NULL) {
      pd->remove() ;
      Page* p = pd->page() ;
      TRY {
	res = p->allocObject(sizeSlot) ;
      } EXCEPTION {
	switch (__YOO_ERR) {
	case YOO_ERR_LOCK :
	case YOO_ERR_UPG :
	  continue ;
	  break ;
	default :
	  UP ;
	}
      }
      // Because it could be reinserted in xlock
      pd->remove() ;
      if (res == NULL) {
	if (i > 0) {
          // Reinsert in the previous entry in case of failed
	  this->_freeList[i-1].insertHead(pd) ;
	}
      } else {
        // In case of success, reinsert 
	this->insert(pd) ;
	return res ;
      }
    }
  }
  return NULL ;
}     


/*= INTERNAL FUNCTION =======================================================*/
ObjectX* Cluster::allocObject(int sizeSlot) 
/*=
[DESCRIPTION :]
  When an object is allocated, it is always in a cluster. It could be default 
  cluster of the class or specified cluster at creation time. Allocation is
  asked to the cluster in case of long or short object. If size asked could not
  fit in a single page, enough pages are allocated from current transaction
  by using segment map. This call initialises the object with oid and size. 
  ClassId if nay must be set by upper layer. Note that size must be inferior to
  the size of a segment. 

  (IN)
    <int sizeSlot>
      => Size in slot of the new object 
  (OUT)
    <ObjectX*>
      => Return virtual pointer on the allocated object. Suppose to always work
      => Trouble are managed in lower layer.

[DIAGNOSTIC :]
[CALL :]
  ObjectX
    => Call in new or changeSize on an object
=*/
{
  ObjectX* res = NULL ;
  PageHdlMem* pd ;
  if (sizeSlot > SLOT_IN_PG) {
    // Compute page size wanted.
    // Allocations of long objects are align on page size
    int numPage = yoo_pageSize(sizeSlot*SLOT_SIZE + PAGE_HEADER_SIZE) ;
    pd = YOO_VirtualServer->allocation(this, numPage) ;
    // Long object has always 0 as object number
    res = pd->page()->object(0) ;
    OID oid = pd->pid() ;
    oid.objNo = 0 ;
    res->initObj(oid, sizeSlot) ;
  } else {
    if (AllocSet[this->getOid()] == NULL) {
      AllocSet += new BlocAllocator(this->getOid()) ;
    } else {
      res = AllocSet[this->getOid()]->alloc(sizeSlot) ;
    }
    if (res == NULL) {
      // Allocate a new page from transaction, make the allocation
      // and insert in blocAllocator
      pd = YOO_VirtualServer->allocation(this, 1) ;
      res = pd->page()->allocObject(sizeSlot) ;
      if (pd->freeSpace() > 0) {
	AllocSet[this->getOid()]->insert(pd) ;
      }
    }
  }
  return res ;
}


/*= STATIC EXTERNAL FUNCTION ================================================*/
void Cluster::reset()
/*=
[DESCRIPTION :]
  Reset the allocators at the end of a transaction if it aborts. 
  The principle of allocation in a cluster is to used space which is on the 
  loaded pages. At the end of a transaction, local cache is flushed (not 
  really because of callback process). The page stored in the allocation list
  could be no more present. So when the current transaction is over, every 
  clusters clear their blocallocator by delete them. The blocallocator will be
  rebuild at the first access to the cluster in the next transaction. 

[DIAGNOSTIC :]
[CALL :]
  LocalTrans
    => Call at the end of the current transaction to reset clusters
=*/
{
  IterClusterSet iter(AllocSet) ;
  for (;iter.more(); iter.next()) {
    delete iter.get() ;
  }

  AllocSet.clear() ;
}


/*= EXTERNAL FUNCTION =======================================================*/
void* Cluster::operator new(size_t s)
/*=
[DESCRIPTION :]
  (IN)  
  (OUT) 
[DIAGNOSTIC :]
[CALL :]
=*/
{
  return ObjectX::operator new(s, YOO_ADMIN_CLU, UNTYPED) ;
}

